@@ -8,7 +8,7 @@ filterwarnings("ignore", category=ConvergenceWarning, message="^Maximum number o
 filterwarnings("ignore", category=ConvergenceWarning, message="^Liblinear failed to converge")
         
 
-################################## LogRegClassifier ##########################################################
+################################## LinearModel ##########################################################
 
 class LinearModel(ModelBase):
     """
@@ -22,7 +22,7 @@ class LinearModel(ModelBase):
         """
         Default model_param
         """
-        model_param = {'verbose':0,}
+        model_param = {}
         return(model_param)
 
     def _init_model(self, model_param=None):
